/*
 *  Part of rvglue source code,
 *  Copyright (C) 2000 Alexander Kroeller (alilein@gmx.de)
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */
#include <stdio.h>
#include <stdlib.h>
#include "rvtypes.h"
#include "mesh.h"

#define MBBOX_SAFETY_DIST 2.0



Mesh::Mesh(RV_PRM_Mesh *m, FILE *f)
{
  data.num_polys=m->num_polys;
  data.num_vecs=m->num_vecs;
  load(f);
}

Mesh::Mesh(RV_W_Mesh *m, FILE *f)
{
  memcpy(&data,m, RV_W_MESH_SIZE);
  load(f);
}

void Mesh::setSurfaceType(rvulong s)
{
  surfacetype=s;
}

rvulong Mesh::surfaceType(void)
{
  return surfacetype;
}

Mesh::Mesh()
{
}

Mesh::~Mesh()
{
  int i;

  for(i=0;i<data.num_polys;i++)
    delete poly[i];

  for(i=0;i<data.num_vecs;i++)
    delete verts[i];

  delete poly;
  delete verts;
}

int Mesh::numPolys(void)
{
  return data.num_polys;
}


Surface *Mesh::createPolySurface(int pidx)
{
  RV_Vertex *pverts[4];
  int i;

  for(i=0;i<poly[pidx]->numverts();i++)
    {
      pverts[i]=&( verts[ poly[pidx]->vertidx(i) ]->data );
    };

  return new Surface(surfacetype,
		     &pverts[0]->pos, &pverts[1]->pos, &pverts[2]->pos,
		     poly[pidx]->numverts()==4 ? &pverts[3]->pos : NULL);
}


void Mesh::setTEColor(rvulong col)
{
  int i;

  for(i=0;i<data.num_polys;i++)
    poly[i]->setTEColor(col);
}


void Mesh::load(FILE *f)
{
  int i;

  poly=new (Polygon*)[data.num_polys];
  verts=new (Vertex*)[data.num_vecs];

  for(i=0;i<data.num_polys;i++)
    poly[i] = new Polygon(f);

  for(i=0;i<data.num_vecs;i++)
    verts[i] = new Vertex(f);

  /* i=ftell(f);
  fseek(f,0,SEEK_END);
  i-=ftell(f);
  if(i!=0)
  printf("MESHLOADDIFF: %d\n",i);*/
}


void Mesh::setTransparency(rvulong t)
{
  int i;

  for(i=0;i<data.num_polys;i++)
    poly[i]->setTransparency(t);
}


void Mesh::translate(rvfloat dx, rvfloat dy, rvfloat dz)
{
  int i;

  for(i=0;i<data.num_vecs;i++)
    verts[i]->translate(dx,dy,dz);
}

void Mesh::rotate(rvfloat mx[3][3])
{
  int i;

  for(i=0;i<data.num_vecs;i++)
    verts[i]->rotate(mx);
}

void Mesh::calc_bounds(void)
{
  rvfloat lo[3], hi[3];
  int i,j;

  lo[0]=lo[1]=lo[2]=1e40;
  hi[0]=hi[1]=hi[2]=-1e40;

  for(i=0;i<data.num_vecs;i++)
    for(j=0;j<3;j++)
      {
	if( verts[i]->data.pos.coord[j] < lo[j] )
	  lo[j]=verts[i]->data.pos.coord[j];
	if( verts[i]->data.pos.coord[j] > hi[j] )
	  hi[j]=verts[i]->data.pos.coord[j];
      };

  data.bbox_xlo = lo[0] - MBBOX_SAFETY_DIST;
  data.bbox_xhi = hi[0] + MBBOX_SAFETY_DIST;
  data.bbox_ylo = lo[1] - MBBOX_SAFETY_DIST;
  data.bbox_yhi = hi[1] + MBBOX_SAFETY_DIST;
  data.bbox_zlo = lo[2] - MBBOX_SAFETY_DIST;
  data.bbox_zhi = hi[2] + MBBOX_SAFETY_DIST;

  for(j=0;j<3;j++)
    data.ball_center.coord[j] = ( lo[j]+hi[j] ) / 2.0;

  data.ball_radius = 
    sqrt(
	 SQR(hi[0]-lo[0])
	 +SQR(hi[1]-lo[1])
	 +SQR(hi[2]-lo[2])
	 )/2.0;
}

void Mesh::write_w(FILE *f)
{
  int i;

  calc_bounds();
  fwrite(&data,RV_W_MESH_SIZE,1,f);

  for(i=0;i<data.num_polys;i++)
    poly[i]->write(f);

  for(i=0;i<data.num_vecs;i++)
    verts[i]->write(f);
}


void Mesh::cleanupPolyList(void)
{
  int cur=0;
  int i;

  while(cur < data.num_polys)
    {
      if(poly[cur] != NULL)
	cur++;
      else
	{
	  for(i=cur;i<data.num_polys-1;i++)
	    poly[i]=poly[i+1];
	  data.num_polys--;
	};
    };
}



void Mesh::deleteTEParts(int ps)
{
  int i;
  bool deleted=false;

  for(i=0;i<data.num_polys;i++)
    if(poly[i]->detTEPart() & ps)
      {
	delete poly[i];
	poly[i]=NULL;
	deleted=true;
      };

  if(deleted)
    cleanupPolyList();
}

  
bool Mesh::isDegenerated(void)
{
  return data.num_polys==0;
}







/*************************************************
void Mesh::add(Polygon p)
{
  if(cursz<maxsz)
    {
      poly[cursz]=new Polygon(p);
      cursz++;
    };
}

Polygon &Mesh::polygon(int i)
{
  return *poly[i];
}

int Mesh::size(void)
{
  return cursz;
}

void Mesh::add(Mesh &grp)
{
  int i;
  
  for(i=0;i<grp.size();i++)
    add(grp.polygon(i));
}


void Mesh::print(FILE *f)
{
  RV_Mesh pg;
  Vector v;
  VertList vl(cursz*4);
  int i;

  crstruct(pg,vl);
  fprintf(f,"Mesh:\n");
  fprintf(f,"  Ball: ");
  v=Vector(&pg.ball_center); v.print(f);
  fprintf(f,", radius %g\n",pg.ball_radius);

  fprintf(f,"  BBox: From ");
  v=Vector(&pg.bound_box[0]); v.print(f);
  fprintf(f,"\n        To   ");
  v=Vector(&pg.bound_box[1]); v.print(f);
  fprintf(f,"\n");
  fprintf(f,"  %d Polys, %d Points\n",pg.num_polys,pg.num_vecs);
  for(i=0;i<cursz;i++)
    {
      printf("Poly #%d\n",i);
      poly[i]->print(f);
    };
}



void Mesh::write(FILE *f)
{
  RV_Mesh pg;
  VertList vlist(cursz*4);
  int i;

  crstruct(pg,vlist);
  fwrite(&pg,RV_POLYGROUP_SIZE,1,f);

  for(i=0;i<cursz;i++)
    poly[i]->write(f,vlist);
  vlist.write(f);
}


void Mesh::write_prm(FILE *f)
{
  RV_Mesh pg;
  RV_PRM_Mesh prm;
  VertList vlist(cursz*4);
  int i;

  crstruct(pg,vlist);

  prm.num_polys=pg.num_polys;
  prm.num_vecs=pg.num_vecs;
  fwrite(&prm,RV_PRM_POLYGROUP_SIZE,1,f);

  for(i=0;i<cursz;i++)
    poly[i]->write(f,vlist);
  vlist.write(f);
}










void Mesh::crstruct(RV_Mesh &g, VertList &vl)
{
  rvfloat xl,xh,yl,yh,zl,zh;
  int v,i;

  xl=yl=zl=1e40;
  xh=yh=zh=-1e40;

  for(i=0;i<cursz;i++)
    for(v=0;v<poly[i]->numverts();v++)
      vl.find(&(poly[i]->point(v)));

  for(i=0;i<vl.size();i++)
    {
      if(vl.vert(i).pos().x() < xl) xl=vl.vert(i).pos().x();
      if(vl.vert(i).pos().y() < yl) yl=vl.vert(i).pos().y();
      if(vl.vert(i).pos().z() < zl) zl=vl.vert(i).pos().z();
      if(vl.vert(i).pos().x() > xh) xh=vl.vert(i).pos().x();
      if(vl.vert(i).pos().y() > yh) yh=vl.vert(i).pos().y();
      if(vl.vert(i).pos().z() > zh) zh=vl.vert(i).pos().z();
    };

  g.bound_box[0].x = xh + SAFETY_DIST;
  g.bound_box[0].y = yh + SAFETY_DIST;
  g.bound_box[0].z = zh + SAFETY_DIST;
  g.bound_box[1].x = xl - SAFETY_DIST;
  g.bound_box[1].y = yl - SAFETY_DIST;
  g.bound_box[1].z = zl - SAFETY_DIST;

  g.num_polys=cursz; g.num_vecs=vl.size();

  g.ball_center.x = (xh+xl)/2.0;
  g.ball_center.y = (yh+yl)/2.0;
  g.ball_center.z = (zh+zl)/2.0;

  g.ball_radius = sqrt(SQR(xh-xl)+SQR(yh-yl)+SQR(zh-zl))/2.0 + SAFETY_DIST;
}
				    *************************************/
