/*
 *  Part of rvglue source code,
 *  Copyright (C) 2000 Alexander Kroeller (alilein@gmx.de)
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */
#include <stdio.h>
#include <stdlib.h>

#include "vector.h"
#include "surface.h"

Surface::Surface(rvulong stype, RV_Vector *v0, RV_Vector *v1, RV_Vector *v2, RV_Vector *v3)
{
  vecs[0]=Vector(v0);
  vecs[1]=Vector(v1);
  vecs[2]=Vector(v2);

  if(v3==NULL)
    nvecs=3;
  else
    {
      nvecs=4;
      vecs[3]=Vector(v3);
    };
  createPolyhedron();
  ph.surface = stype;
}

Vector Surface::planarNormal(void)
{
  Vector v0, v2;

  v0 = vecs[0] - vecs[1];
  v2 = vecs[2] - vecs[0];
  return Vector(v0,v2);
}


void Surface::createPlane(RV_Plane *pl, Vector &normal, Vector &pos)
{
  memcpy(&pl->normal, &normal.data, RV_VECTOR_SIZE);
  pl->distance = -pos.scalarprod(normal);
}


void Surface::write(FILE *f)
{
  fwrite(&ph, RV_POLYHEDRON_SIZE, 1, f);
}

void Surface::expandGrid(LookupGrid *grid)
{
  int i;

  for(i=0;i<nvecs;i++)
    grid->expandGrid(vecs[i]);
}

void Surface::createPolyhedron(void)
{
  Vector normal, snormal, tmpvec;
  int i;

  ph.type   = (nvecs == 4) ? 1 : 0;
  ph.surface= 0;


  /******* BOUNDING BOX *********/
  ph.bbox_xlo = ph.bbox_ylo = ph.bbox_zlo = 1.0e40;
  ph.bbox_xhi = ph.bbox_yhi = ph.bbox_zhi = -1.0e40;
  
  for(i=0;i<nvecs;i++)
    {
      if(vecs[i].x() < ph.bbox_xlo) ph.bbox_xlo=vecs[i].x();
      if(vecs[i].y() < ph.bbox_ylo) ph.bbox_ylo=vecs[i].y();
      if(vecs[i].z() < ph.bbox_zlo) ph.bbox_zlo=vecs[i].z();

      if(vecs[i].x() > ph.bbox_xhi) ph.bbox_xhi=vecs[i].x();
      if(vecs[i].y() > ph.bbox_yhi) ph.bbox_yhi=vecs[i].y();
      if(vecs[i].z() > ph.bbox_zhi) ph.bbox_zhi=vecs[i].z();
    };

  ph.bbox_xlo -= 1.0;
  ph.bbox_ylo -= 1.0;
  ph.bbox_zlo -= 1.0;

  ph.bbox_xhi += 1.0;
  ph.bbox_yhi += 1.0;
  ph.bbox_zhi += 1.0;
  


  /********* MASTER PLANE *******/
  normal=planarNormal(); /* normal on surface */
  createPlane(&ph.plane[0], normal, vecs[1]);


  /********* BOUNDING PLANES ******/
  for(i=0;i<nvecs;i++)
    {
      tmpvec = vecs[(i+1) % nvecs] - vecs[i];
      snormal=Vector(normal, tmpvec);
      createPlane(&ph.plane[i+1], snormal, vecs[i]);
    };
}



void Surface::enterIntoGrid(LookupGrid *grid, rvulong myidx)
{
  int xlo, zlo, xhi, zhi;
  int x,z;

  xlo=grid->xgridcoord(ph.bbox_xlo);
  zlo=grid->zgridcoord(ph.bbox_zlo);
  xhi=grid->xgridcoord(ph.bbox_xhi + 1.0);
  zhi=grid->zgridcoord(ph.bbox_zhi + 1.0);
  
  if(xlo<0) xlo=0;
  if(zlo<0) zlo=0;

  for(x=xlo;x<=xhi;x++)
    for(z=zlo;z<=zhi;z++)
      grid->lichain(x,z)->add(myidx);
}
